import json
import os
from collections import UserString

KEYS_PATH = os.path.expanduser("~/.local/share/wuenlp-tools/keys.json")


def read_keys():
    """Read keys from the JSON file. Returns an empty dict if the file doesn't exist or is corrupted."""
    if os.path.isfile(KEYS_PATH):
        try:
            with open(KEYS_PATH, "r") as f:
                return json.load(f)
        except json.JSONDecodeError:
            return {}
    return {}


def write_keys(keys):
    """Write the keys dictionary to the JSON file."""
    os.makedirs(os.path.dirname(KEYS_PATH), exist_ok=True)
    with open(KEYS_PATH, "w") as f:
        json.dump(keys, f, indent=2)


def get_api_key(key_name):
    """
    Get an API key by name from the file.
    If not present, prompt the user and store it.
    """
    keys = read_keys()
    if key_name in keys:
        return keys[key_name]
    else:
        key_value = input(f"Please enter the API key for {key_name}: ").strip()
        keys[key_name] = key_value
        write_keys(keys)
        return key_value


class LazyKey(UserString):
    """
    A lazy-loading key that behaves like a string using UserString.

    The actual API key is only requested (and cached) when the underlying data is accessed.
    """

    def __init__(self, key_name):
        self.key_name = key_name
        self._loaded = False
        self._data = None  # _data will hold the actual key value
        # We intentionally initialize the UserString with an empty string.
        super().__init__("")

    @property
    def data(self):
        """Load and cache the data on first access."""
        if not self._loaded:
            self._data = get_api_key(self.key_name)
            self._loaded = True
        return self._data

    @data.setter
    def data(self, value):
        self._data = value
        self._loaded = True

    def __str__(self):
        return self.data

    def __repr__(self):
        return repr(self.data)

    # UserString uses self.data for many of its operations, so we simply return the loaded data.
    def __add__(self, other):
        return self.data + other

    def __radd__(self, other):
        return other + self.data
