from __future__ import annotations

import json

import atexit
from dataclasses import dataclass
from functools import lru_cache
from typing import T

import openai
import requests
from loguru import logger
from openai import NOT_GIVEN
from pydantic import BaseModel

from .model_cost import get_cost

from wuenlp_tools.keys import OPENAI_API_KEY, OPENROUTER_API_KEY

openai.api_key = OPENAI_API_KEY

RUN_COST = 0


@atexit.register
def print_run_cost():
    logger.info(f"Total cost of the run: {RUN_COST:.2f} €")


@dataclass
class LLMArchitecture:
    name: str
    model: str
    max_tokens: int
    openai: bool
    openrouter: bool = False

    @property
    def estimated_num_characters(self):
        return int(self.max_tokens * 3.5)


llama31_70b = LLMArchitecture(name="llama3.1:70b", model="llama3.1:70b", max_tokens=-1, openai=False)
llama31_8b = LLMArchitecture(name="llama3.1:8b", model="llama3.1:8b", max_tokens=-1, openai=False)
llama32_7b = LLMArchitecture(name="llama3.2:7b", model="llama3.1:8b", max_tokens=-1, openai=False)
# gpt4o = LLMArchitecture(name="gpt-4o", model="gpt-4o", max_tokens=128000, openai=True)
gpt4mini = LLMArchitecture(name="gpt-4o-mini", model="gpt-4o-mini", max_tokens=128000, openai=True)
gpt4omini = gpt4mini

o3mini = LLMArchitecture(name="o3-mini", model="o3-mini", max_tokens=128000, openai=True)

r1 = LLMArchitecture(name="r1", model="deepseek/deepseek-r1", max_tokens=128000, openai=False, openrouter=True)


class Language(BaseModel):
    name: str


class LLM:
    def __init__(self, model: LLMArchitecture, system_prompt: str, post_prompt: str = None, cache_maxsize: int = 0,
                 seed=NOT_GIVEN, json: bool = False, output_format: T[BaseModel] = None,
                 base_url: str = "https://ollama.professor-x.de/v1/"):
        if model.openai:
            self.base_url = None
            openai.api_key = OPENAI_API_KEY
        elif model.openrouter:
            self.base_url = "https://openrouter.ai/api/v1/"
            openai.api_key = OPENROUTER_API_KEY
        else:
            self.base_url = base_url
        self.model = model
        self.system_prompt = system_prompt
        self.post_prompt = ("...\n\n" + post_prompt) if post_prompt else ""
        self.cache_maxsize = cache_maxsize
        self.seed = seed
        self.json = json
        self.output_format = output_format

        self.__call__ = lru_cache(maxsize=self.cache_maxsize)(self.__call__)

    def __call__(self, prompt) -> str | T:
        global RUN_COST

        if self.base_url:
            openai.base_url = self.base_url
        try:
            model_max_chars = self.model.estimated_num_characters
            if model_max_chars > -1 and len(prompt) > model_max_chars:
                prompt = prompt[:model_max_chars]
                logger.warning(f"Truncated prompt to {model_max_chars} characters.")
            prompt += self.post_prompt
            logger.debug(f"System prompt is: {self.system_prompt}")
            shortened_prompt = prompt if len(prompt) < 1000 else prompt[:1000] + '\n...\n' + prompt[-1000:]
            logger.debug(f"Querying {self.model}, seed {self.seed} with prompt: {shortened_prompt}")
            system_role = "system" if not self.model.name.startswith("o3") else "developer"
            response = openai.beta.chat.completions.parse(model=self.model.model,
                                                          messages=[
                                                              {"role": system_role, "content": self.system_prompt},
                                                              {"role": "user",
                                                               "content": prompt}], seed=self.seed,
                                                          response_format=self.output_format if self.output_format is not None else {
                                                              "type": "json_object"} if self.json else {
                                                              "type": "text"},
                                                          extra_headers={
                                                              "Authorization": f"Bearer {openai.api_key}"
                                                          },
                                                          extra_body={
                                                              "include_reasoning": True
                                                          } if self.model.openrouter else None,
                                                          )

            if self.output_format is not None:
                content = response.choices[0].message.parsed
                if content is None:
                    try:
                        content = json.loads(response.choices[0].message.content)
                    except json.JSONDecodeError:
                        content = response.choices[0].message.content
            else:
                content = response.choices[0].message.content
            if self.model.openai:
                cost = get_cost(response, self.model.model)
                RUN_COST += cost
            elif self.model.openrouter:
                url = f"https://openrouter.ai/api/v1/generation?id={response.id}"
                gen_response = requests.post(
                    url=url,
                    headers={
                        "Authorization": f"Bearer {OPENROUTER_API_KEY}",
                        "Content-Type": "application/json"
                    }).json()

                try:
                    cost = gen_response["total_cost"]
                except KeyError:
                    cost = 0
            else:
                cost = 0

            try:
                response_reasoning = response.choices[0].message.reasoning
            except Exception as e:
                response_reasoning = None

            logger.debug(f"Model {self.model} says: {content}")
            if response_reasoning:
                logger.debug(f"Reasoning: {response_reasoning}")
            logger.debug(f"Cost: {cost:.2f} €")
            logger.debug(f"Total cost: {RUN_COST:.2f} €")
            return content
        finally:
            openai.base_url = None


class YesNoReason(BaseModel):
    answer: bool
    reason: str


class YesNo(BaseModel):
    answer: bool


class YesNoLLM(LLM):
    def __init__(self, model: LLMArchitecture, system_prompt: str, post_prompt: str = None, cache_maxsize: int = 0,
                 seed=NOT_GIVEN, max_tries=3, reason: bool = False):
        if model.openai:
            super().__init__(model, system_prompt, post_prompt, cache_maxsize, seed=seed,
                             output_format=YesNoReason if reason else YesNo)
        else:
            if not post_prompt or "yes" not in post_prompt:
                logger.warning("Post prompt does not contain 'yes'. Adding instruction.")
                post_prompt = post_prompt + "Answer only with yes or no, nothing else." if not reason else "Answer with yes or no, followed by a reason"
            super(YesNoLLM, self).__init__(model, system_prompt,
                                           post_prompt,
                                           cache_maxsize,
                                           seed=seed, json=False)
        self.max_tries = max_tries
        self.reason = reason

    def __call__(self, prompt) -> (bool, str):
        if self.model.openai:
            response = super().__call__(prompt)
            return response
        tries = 0
        while tries < self.max_tries:
            response = super().__call__(prompt)
            response_lower = response.lower().strip("*").strip()
            if response_lower.startswith("yes") or response_lower.startswith("no"):
                answer = response_lower.startswith("yes")
                if self.reason:
                    return YesNoReason(answer=answer, reason=response)
                else:
                    return YesNo(answer=answer)
            tries += 1
        raise ValueError(f"Could not get a yes/no answer after {self.max_tries} tries.")


# @cache_to_disk(maxsize=-1)
def get_example_for_schema(schema: BaseModel, query: str):
    language_llm = LLM(gpt4mini, system_prompt="Return the English name of the language of the provided text.", seed=42,
                       output_format=Language)

    language = language_llm(query).name

    system_prompt = f"Your job is to guide a language model in generating a response in the provided output format. Your instructions should be in {language}. Address the language model informally."
    instruction_gpt = LLM(gpt4mini, system_prompt, seed=42)
    example_gpt = LLM(gpt4mini, system_prompt, seed=42, output_format=schema)

    example: schema = example_gpt("Give an example of the output format you want the model to generate.")
    return instruction_gpt(
        "Start by telling the model 'Your output should match this format:'.") + "\n" + example.json()


if __name__ == '__main__':
    gpt = YesNoLLM(llama31_70b, "This is a test prompt.", seed=42, reason=True)
    response = gpt("What is the meaning of life?")
    print(response)
