from typing import Type

import numpy as np
import pandas as pd
import plotly.express as px
from loguru import logger

from wuenlp.impl.UIMANLPStructs import UIMADocument, UIMAAnnotation, UIMASystemScene


# Function to format text with newlines
from wuenlp_tools.utils.summarize import summarize_doc


def format_text(text, line_length=20, max_lines=75):
    # Split long texts into lines
    split = text.split(" ")
    lines = [' '.join(split[i:i + line_length]) for i in range(0, len(split), line_length)]

    if len(lines) > max_lines:
        lines = lines[:max_lines // 2] + [f"[{len(lines) - max_lines} more lines]"] + lines[-max_lines // 2:]
    return "<br>".join(lines)


def plot_sentiment_interactive(doc: UIMADocument, segment_type: Type[UIMAAnnotation] = UIMASystemScene, smooth: int = 0,
                               line_length=20,
                               summary: bool = False, call_summarizer_if_necessary: bool = False):
    sentiments = []
    texts = []
    text_lengths = []
    keywords = []

    segments = doc._get_annos_of_type(segment_type)
    if not segments:
        raise ValueError(f"No {segment_type.__name__} found in {doc.path.stem}")

    if call_summarizer_if_necessary and not segments[0].additional_features.get('llama_summary', None):
        doc = summarize_doc(doc, segment_type=segment_type)

    # Extracting scenes (assuming 'document' is UIMADocument and has a method to get system scenes)
    for segment in segments:
        llama_sentiment = segment.additional_features.get('llama_sentiment', None)
        if llama_sentiment is not None:
            sentiments.append(llama_sentiment)
            text = segment.additional_features["llama_summary"] if summary else segment.text
            formatted_text = format_text(text, line_length)
            texts.append(formatted_text)
            keyword = "<br>".join(segment.additional_features.get('llama_keywords', []))
            logger.info(f"Keywords: {keyword}")
            keywords.append(keyword)
            text_lengths.append(len(segment.text))  # Record the length of the original text

    if smooth:
        smoothed_sentiments = []
        num_sentiments = len(sentiments)
        for i in range(num_sentiments):
            smoothed_sentiments.append(sum(sentiments[max(0, i - smooth):min(num_sentiments, i + smooth + 1)]) / (
                    min(num_sentiments, i + smooth + 1) - max(0, i - smooth)))
        sentiments = smoothed_sentiments

    # Create a DataFrame for Plotly
    df = pd.DataFrame({'Sentiment': sentiments, 'Scene': texts, 'Text Length': text_lengths, 'Keywords': keywords})

    # Calculate the mean and standard deviation
    mean_sentiment = df['Sentiment'].mean()
    std_sentiment = df['Sentiment'].std()

    # Define threshold for outliers (for example: mean ± 1.5 * std deviation)
    threshold_upper = mean_sentiment + 1.5 * std_sentiment
    threshold_lower = mean_sentiment - 1.5 * std_sentiment

    size_scale = 1000  # Adjust this scaling factor for marker sizes
    df['Marker Size'] = np.array(df['Text Length']) / size_scale

    # Create an interactive plot
    fig = px.line(df, x=df.index, y='Sentiment',
                  hover_data=['Scene'],
                  labels={'Sentiment': 'Llama Sentiment', 'Scene': 'Scene Text'},
                  title=f'Sentiment per Segment in {doc.path.stem}{" (Scenes)" if segments else "Chunks"}{" (Smoothed)" if smooth else ""}')

    fig.update_traces(line=dict(width=2),
                      hovertemplate='%{hovertext}<extra></extra>',
                      mode='lines+markers+text',
                      marker=dict(size=df['Marker Size'], opacity=0.7),
                      hovertext=df['Scene'],
                      )

    for i in range(len(df)):
        if df['Sentiment'].iloc[i] > threshold_upper:
            # High outlier annotation
            fig.add_annotation(
                x=df.index[i],
                y=df['Sentiment'].iloc[i] + 0.01,  # Adjust for space above
                text=df['Keywords'].iloc[i],
                showarrow=True,
                arrowhead=2,
                ax=0,
                ay=-30,  # Arrow pointing downwards
                textangle=0,  # Straight text for high points
                font=dict(size=10)
            )
        elif df['Sentiment'].iloc[i] < threshold_lower:
            # Low outlier annotation
            fig.add_annotation(
                x=df.index[i],
                y=df['Sentiment'].iloc[i] - 0.01,  # Adjust for space below
                text=df['Keywords'].iloc[i],
                showarrow=True,
                arrowhead=2,
                ax=0,
                ay=30,  # Arrow pointing upwards
                textangle=0,  # Straight text for low points
                font=dict(size=10)
            )

    fig.show()
